<?php declare(strict_types=1);

namespace MyPrayerJournal;

class Layout
{
    /**
     * Create the `DOCTYPE` declaration, `html`, and `head` tags for the page
     *
     * @param string $title The title of the page
     */
    public static function htmlHead(string $title): void
    {
        if (is_htmx()) {
            echo "<!DOCTYPE html><html lang=en><head lang=en><title>$title &#xab; myPrayerJournal</title></head>";
        } else {
            echo <<<HEAD
                <!DOCTYPE html>
                <html lang=en>
                <head>
                    <meta name=viewport content="width=device-width, initial-scale=1">
                    <meta name=description content="Online prayer journal - free w/Google or Microsoft account">
                    <title>$title &#xab; myPrayerJournal</title>
                    <link href=https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css rel=stylesheet
                          integrity="sha384-T3c6CoIi6uLrA9TneNEoa7RxnatzjcDSCmG1MXxSR1GAsXEV/Dwwykc2MPK8M2HN"
                          crossorigin=anonymous>
                    <link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel=stylesheet>
                    <link href=/style/style.css rel=stylesheet>
                </head>
                HEAD;
        }
    }

    private static function navLink(string $url, string $text): void
    {
        $classAttr = match (true) {
            str_starts_with($_SERVER['PHP_SELF'], $url) => ['class' => 'is-active-route'],
            default => []
        };
        echo '<li class=nav-item>';
        page_link($url, $text, $classAttr);
    }

    /**
     * The default navigation bar, which will load the items on page load, and whenever a refresh event occurs
     */
    public static function navBar(): void
    { ?>
        <nav class="navbar navbar-dark" role="navigation">
            <div class=container-fluid><?php
                page_link('/', '<span class=m>my</span><span class=p>Prayer</span><span class=j>Journal</span>',
                    ['class' => 'navbar-brand']); ?>
                <ul class="navbar-nav me-auto d-flex flex-row"><?php
                    if (key_exists('user_id', $_SESSION)) {
                        self::navLink('/journal', 'Journal');
                        self::navLink('/requests/active', 'Active');
                    if (key_exists('has_snoozed', $_SESSION)) self::navLink('/requests/snoozed', 'Snoozed');
                        self::navLink('/requests/answered', 'Answered'); ?>
                        <li class=nav-item><a href=/user/log-off>Log Off</a><?php
                    } else { ?>
                        <li class=nav-item><a href=/user/log-on>Log On</a><?php
                    }
                    self::navLink('/docs', 'Docs'); ?>
                </ul>
            </div>
        </nav><?php
    }

    /**
     * Drop .0 or .0.0 from the end of the version to format it for display
     *
     * @return string The version of the application for user display
     */
    private static function displayVersion(): string {
        [$major, $minor, $rev] = explode('.', MPJ_VERSION);
        $minor = $minor == '0' ? '' : ".$minor";
        $rev = match (true) {
            $rev == '0'                 => '',
            str_starts_with($rev, '0-') => substr($rev, 1),
            default                     => ".$rev"
        };
        return "v$major$minor$rev";
    }

    /**
     * Create the footer
     */
    public static function htmlFoot(): void
    { ?>
        <footer class=container-fluid>
            <p class="text-muted text-end">
                myPrayerJournal <?=self::displayVersion();?><br>
                <em><small><?php
                    page_link('/legal/privacy-policy', 'Privacy Policy');
                    echo ' &bull; ';
                    page_link('/legal/terms-of-service', 'Terms of Service');
                    echo ' &bull; '; ?>
                    <a href=https://git.bitbadger.solutions/bit-badger/myPrayerJournal target=_blank
                       rel=noopener>Developed</a> and hosted by
                    <a href=https://bitbadger.solutions target=_blank rel=noopener>Bit Badger Solutions</a>
                </small></em>
            <script src=https://unpkg.com/htmx.org@2.0.0 crossorigin=anonymous
                    integrity="sha384-wS5l5IKJBvK6sPTKa2WZ1js3d947pvWXbPJ1OmWfEuxLgeHcEbjUUA5i9V5ZkpCw"></script>
            <script>if (!htmx) document.write('<script src=\"/script/htmx.min.js\"><\/script>')</script>
            <script async src=https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/js/bootstrap.bundle.min.js
                    integrity="sha384-C6RzsynM9kWDrMNeT87bh95OGNyZPhcTNXj1NW7RuBCsyN/o0jlpcV8Qyq46cDfL"
                    crossorigin=anonymous></script>
            <script>
                setTimeout(function () {
                    if (!bootstrap) document.write('<script src=\"/script/bootstrap.bundle.min.js\"><\/script>')
                }, 2000)
            </script>
            <script src=/script/mpj.js></script>
        </footer><?php
    }

    /// Create a card when there are no results found
    public static function noResults(string $heading, string $link, string $buttonText, string $text): void
    { ?>
        <div class=card>
            <h5 class=card-header><?=$heading?></h5>
            <div class="card-body text-center">
                <p class=card-text><?=$text?></p><?php
                page_link($link, $buttonText, ['class' => 'btn btn-primary']); ?>
            </div>
        </div><?php
    }
}
