<?php declare(strict_types=1);

namespace MyPrayerJournal;

use BitBadger\PDODocument\{DocumentException, Find};
use Exception;
use JsonSerializable;
use Visus\Cuid2\Cuid2;

/**
 * A prayer request
 */
class Request implements JsonSerializable
{
    /**
     * @param string $id The ID for the request
     * @param string $enteredOn The date/time this request was originally entered
     * @param string $userId The ID of the user to whom this request belongs
     * @param string|null $snoozedUntil The date/time the snooze expires for this request (null = not snoozed)
     * @param string|null $showAfter The date/time the current recurrence period is over (null = immediate)
     * @param Recurrence $recurrence The recurrence for this request
     * @param History[] $history The history of this request
     * @param Note[] $notes Notes regarding this request
     * @throws Exception If the ID generation fails
     */
    public function __construct(public string $id = '', public string $enteredOn = '', public string $userId = '',
                                public ?string $snoozedUntil = null, public ?string $showAfter = null,
                                public Recurrence $recurrence = new Recurrence(RecurrencePeriod::Immediate),
                                public array $history = [], public array $notes = [])
    {
        if ($id == '') {
            $this->id = (new Cuid2())->toString();
        }
    }

    /**
     * Find a request by its ID
     *
     * @param string $id The ID of the request
     * @return Request|false The request if it is found and belongs to the current user, false if not
     * @throws DocumentException If any is encountered
     */
    public static function byId(string $id): Request|false
    {
        $req = Find::byId(Table::REQUEST, $id, self::class);
        return ($req && $req->userId == $_SESSION['user_id']) ? $req : false;
    }

    public function jsonSerialize(): mixed
    {
        $values = [
            'id'         => $this->id,
            'enteredOn'  => $this->enteredOn,
            'userId'     => $this->userId,
            'recurrence' => $this->recurrence,
            'history'    => $this->history,
            'notes'      => $this->notes
        ];
        if (!is_null($this->snoozedUntil)) $values['snoozedUntil'] = $this->snoozedUntil;
        if (!is_null($this->showAfter))    $values['showAfter']    = $this->showAfter;
        return $values;
    }
}
